@@ -13,6 +13,7 @@ from marshmallow import Schema, EXCLUDE
 from time import sleep
 
 import fusion_platform
+from fusion_platform.common.utilities import value_from_read_only
 from fusion_platform.models import fields
 from fusion_platform.models.data import Data
 from fusion_platform.models.model import Model, ModelError
@@ -103,11 +104,16 @@ class ProcessExecutionStatusSchema(Schema):
     Nested schema class for execution status.
     """
     process_execution_id = fields.UUID(required=True)
+    group_id = fields.UUID(allow_none=True)
+    group_index = fields.Integer(allow_none=True)
+    group_count = fields.Integer(allow_none=True)
     job_status = fields.String(required=True)
+    abort_reason = fields.String(allow_none=True)
     started_at = fields.DateTime(required=True)
     ended_at = fields.DateTime(allow_none=True)
     delete_warning_status = fields.String(required=True)
     delete_expiry = fields.DateTime(required=True)
+    notified = fields.Boolean(allow_none=True)
 
     class Meta:
         """
@@ -233,7 +239,8 @@ class ProcessSchema(Schema):
     price = fields.Decimal(required=True, metadata={'read_only': True})  # Changed to prevent this being updated.
 
     deletable = fields.String(allow_none=True, metadata={'read_only': True})  # Changed to prevent this being updated.
-    has_executions = fields.Boolean(required=True, metadata={'hide': True})  # Changed to hide as an attribute.
+    has_executions = fields.String(required=True, metadata={'hide': True})  # Changed to hide as an attribute.
+    has_executions_organisation_id = fields.String(required=True, metadata={'hide': True})  # Changed to hide as an attribute.
     executions = fields.List(fields.Nested(ProcessExecutionStatusSchema()), allow_none=True, metadata={'hide': True})  # Changed to hide as an attribute.
 
     # Removed creator.
@@ -263,6 +270,7 @@ class Process(Model):
     _PATH_BASE = f"{_PATH_ROOT}/{{process_id}}"
 
     # Override the standard model paths.
+    _PATH_COPY = f"{_PATH_BASE}/copy"
     _PATH_CREATE = _PATH_ROOT
     _PATH_DELETE = _PATH_BASE
     _PATH_GET = _PATH_BASE
@@ -278,7 +286,7 @@ class Process(Model):
     _PROCESS_STATUS_EXECUTE = 'execute'
 
     # The maximum number of seconds to wait after an execution was meant to start.
-    _EXECUTE_WAIT_TOLERANCE = 60
+    _EXECUTE_WAIT_TOLERANCE = 900
 
     # Validation parsing for constrained values.
     _VALIDATION_DELIMITER = ';'
@@ -326,6 +334,62 @@ class Process(Model):
             model = OptionDataTypeSchema().load({data_type: value})
             return model.get(data_type)
 
+    def copy(self, name):
+        """
+        Creates a new template process from an existing process object. This process is not persisted to the Fusion Platform<sup>&reg;</sup>.
+
+        Args:
+            name: The name of the copy.
+
+        Returns:
+            The new template process object.
+
+        Raises:
+            RequestError: if the copy fails.
+            ModelError: if the model could not be created and validated by the Fusion Platform<sup>&reg;</sup>.
+        """
+        # Get a new template for the process model using the process as the source of the copy.
+        process = Process(self._session)
+
+        # Copy the process model.
+        response = self._session.request(path=self._get_path(self.__class__._PATH_COPY), method=Session.METHOD_GET)
+
+        # Assume that the resulting model is held within the expected key within the resulting dictionary.
+        if Model._RESPONSE_KEY_MODEL not in response:
+            raise ModelError(i18n.t('models.process.failed_copy'))
+
+        # Make sure we have an organisation id.
+        modified_response = response.get(Model._RESPONSE_KEY_MODEL, {})
+        modified_response['organisation_id'] = self.organisation_id
+
+        # Load the response into the model. We ignore missing required fields and those which are None.
+        process._set_model_from_response(modified_response, partial=True)
+
+        # Update the options because these are not copied across to the new model, even though they are returned in the response.
+        options = modified_response.get('options', [])
+        options = [] if options is None else options
+
+        for option in options:
+            # Set the option value. The value may be a string, rather than the required type, but that will automatically be converted.
+            if (option.get(self.__class__._FIELD_NAME) is not None) and (option.get(self.__class__._FIELD_VALUE) is not None):
+                process.__set_option(name=option.get(self.__class__._FIELD_NAME), value=option.get(self.__class__._FIELD_VALUE), coerce_value=True)
+
+        # And now the same for the inputs. Here we also validate each input by retrieving it.
+        inputs = modified_response.get('inputs', [])
+        inputs = [] if inputs is None else inputs
+
+        for i, input in enumerate(inputs):
+            # See if we can find the associated model.
+            if input.get(self.__class__._FIELD_ID) is not None:
+                data = Data._model_from_api_id(self._session, organisation_id=self.organisation_id, data_id=input.get(self.__class__._FIELD_ID))
+                process.__set_input(number=i + 1, data=data)
+
+        # Set the name.
+        process.update(name=name)
+
+        # Return the copy.
+        return process
+
     def create(self):
         """
         Attempts to persist the template process in the Fusion Platform<sup>&reg;</sup> so that it can be executed.
@@ -471,7 +535,7 @@ class Process(Model):
         """
         for input in self._model.get(self.__class__._FIELD_INPUTS, []):
             # We first have to remove the mapping proxy so that we can wrap the dictionary in a model.
-            input = dict(input)
+            input = value_from_read_only(input)
 
             # See if we can find the associated model.
             if input.get(self.__class__._FIELD_ID) is not None:
@@ -494,7 +558,7 @@ class Process(Model):
         """
         for option in self._model.get(self.__class__._FIELD_OPTIONS, []):
             # We first have to remove the mapping proxy so that we can wrap the dictionary in a model.
-            option = dict(option)
+            option = value_from_read_only(option)
 
             # If the option is a constrained data type, then add in the constrained names and values from the validation.
             if option.get(self.__class__._FIELD_DATA_TYPE) == fusion_platform.DATA_TYPE_CONSTRAINED:
@@ -574,7 +638,7 @@ class Process(Model):
         # We can now update the input.
         self._set_field([self.__class__._FIELD_INPUTS, index, self.__class__._FIELD_ID], data.id)
 
-    def __set_option(self, name=None, option=None, value=None):
+    def __set_option(self, name=None, option=None, value=None, coerce_value=False):
         """
         Sets the specified option for the process to the value. An exception is raised if the process is in the execute status, the option does not exist or the
         value has the wrong type.
@@ -583,6 +647,7 @@ class Process(Model):
             name: The option name to set. Either the name or the option must be provided.
             option: The option object for the option to set. Either the name or the option must be provided.
             value: The value for the option.
+            coerce_value: Optionally coerce the supplied value to be the correct type. Default False.
 
         Raises:
             ModelError: if the process is the execute status.
@@ -615,6 +680,11 @@ class Process(Model):
         # Check that the option has the same data type as the value. We cannot check this if either value is None. Note that supplied ints can be used for floats.
         data_type = found_option.get(self.__class__._FIELD_DATA_TYPE)
         existing_value = self.__class__.__coerce_value(found_option[self.__class__._FIELD_VALUE], data_type)
+
+        # Optionally coerce the supplied value, now that we know its data type.
+        if coerce_value:
+            value = self.__class__.__coerce_value(value, data_type)
+
         class_matches = isinstance(value, existing_value.__class__)
 
         if isinstance(value, int) and isinstance(existing_value, float):
@@ -637,7 +707,7 @@ class Process(Model):
         # Send the request and load the resulting model.
         self._send_and_load(self._get_path(self.__class__._PATH_STOP), method=Session.METHOD_POST)
 
-    def update(self, input_number=None, input=None, data=None, option_name=None, option=None, value=None, **kwargs):
+    def update(self, input_number=None, input=None, data=None, option_name=None, option=None, value=None, coerce_value=False, **kwargs):
         """
         Attempts to update the model object with the given values. This assumes the model is updated using a PATCH RESTful request. This assumes that the patch body
         contains key names which include the name of the model class. Overridden to prevent changes if the process is being executed and to handle the special cases
@@ -650,6 +720,7 @@ class Process(Model):
             option_name: The option name to set. Either the name or the option must be provided when setting an option.
             option: The option object for the option to set. Either the name or the option must be provided when setting an option.
             value: The value for the option.
+            coerce_value: Optionally coerce the supplied value to be the correct type. Default False.
             kwargs: The model attributes which are to be patched.
 
         Raises:
@@ -668,7 +739,7 @@ class Process(Model):
 
         # Deal with the special case of options.
         if (option_name is not None) or (option is not None):
-            self.__set_option(name=option_name, option=option, value=value)
+            self.__set_option(name=option_name, option=option, value=value, coerce_value=coerce_value)
 
         # Now update the model, persisting as needed.
         super(Process, self).update(**kwargs)
@@ -698,11 +769,10 @@ class Process(Model):
     def wait_for_next_execution(self):
         """
         Waits for the next execution of the process to start, according to its schedule. If executions are already started, this method will return immediately.
-        Otherwise, this method will block until the next scheduled execution has started. An exception will be raised if the process is not being executed.
+        Otherwise, this method will block until the next scheduled execution has started.
 
         Raises:
             RequestError: if any request fails.
-            ModelError: if the process is not being executed.
             ModelError: if a model could not be loaded or validated from the Fusion Platform<sup>&reg;</sup>.
         """
         # Wait until we find the next execution.
@@ -710,22 +780,28 @@ class Process(Model):
             # Load in the most recent version of the model.
             self.get(organisation_id=self.organisation_id)
 
-            # Make sure the process is executable.
-            if self.process_status != Process._PROCESS_STATUS_EXECUTE:
-                raise ModelError(i18n.t('models.process.not_executable'))
-
-            # Get the most recent execution for the process. This assumes that the executions are returned with the most recent first.
+            # Check whether we have the next execution listed in the model. We sort the list to find the most recent.
             self._logger.debug('checking for next execution')
-            executions = ProcessExecution._models_from_api_path(self._session, self._get_path(self.__class__._PATH_EXECUTIONS), items_per_request=1, reverse=True)
-            execution = next(executions, None)
+            executions = []
+
+            if self._model.get(self.__class__._FIELD_HAS_EXECUTIONS, False):
+                executions = [item for item in self._model.get(self.__class__._FIELD_EXECUTIONS, [])]  # Turn the read-only field into a list we can sort.
+                executions = sorted(executions, key=lambda item: item.get(self.__class__._FIELD_STARTED_AT), reverse=True)  # Sort by most recent first.
+
+            execution = executions[0] if len(executions) > 0 else None
+
+            # If the execution is in a group, then make sure that all the executions in the group have started.
+            if (execution is not None) and (execution.get(self.__class__._FIELD_GROUP_ID) is not None):
+                group = [item for item in executions if item.get(self.__class__._FIELD_GROUP_ID) == execution.get(self.__class__._FIELD_GROUP_ID)]
+                execution = None if (len(group) < execution.get(self.__class__._FIELD_GROUP_COUNT)) else execution
 
             # Ignore any execution older than when the next execution is expected. This assumes that the process repeat start is maintained correctly, and that it
             # is set after any corresponding executions have been created for the current repeat start.
-            execution = None if (execution is not None) and (execution.created_at < self.repeat_start) else execution
+            execution = None if (execution is not None) and (execution.get(self.__class__._FIELD_STARTED_AT) < self.repeat_start) else execution
 
             # Stop if we have an execution which is beyond the repeat start date.
             if execution is not None:
-                self._logger.debug('execution %s found', execution.id)
+                self._logger.debug('execution %s found', execution.get(self.__class__._FIELD_ID))
                 break
 
             # If we have no recent executions, and longer than the allowed period has elapsed since the next execution was meant to start, then raise an exception.
